/*
 * qstat 2.0a
 * by Steve Jankowski
 * steve@activesw.com
 * http://www.activesw.com/people/steve/qstat.html
 *
 * Thanks to Per Hammer for the OS/2 patches (per@mindbend.demon.co.uk)
 *
 * Inspired by QuakePing by Len Norton
 *
 * Copyright 1996,1997 by Steve Jankowski
 *
 *   Permission granted to use this software for any purpose you desire
 *   provided that existing copyright notices are retained verbatim in all
 *   copies and derived works.
 */

#define VERSION "2.0b"

/* OS/2 defines */
#ifdef __OS2__
#define BSD_SELECT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <string.h>
#include <ctype.h>

#ifdef unix
#define _ISUNIX
#endif
#ifdef __unix
#define _ISUNIX
#endif

#ifdef _ISUNIX
#include <unistd.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>

#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#ifndef INADDR_NONE
#define INADDR_NONE ~0
#endif
#define sockerr()	errno
#endif /* _ISUNIX */

#ifdef __hpux
extern int h_errno;
#endif

#ifdef _WIN32
/* #define FD_SETSIZE 100 */
#include <winsock.h>
#include <sys/timeb.h>
#define close(a) closesocket(a)
int gettimeofday(struct timeval *now, void *blah)
{
    struct timeb timeb;
    ftime( &timeb);
    now->tv_sec= timeb.time;
    now->tv_usec= (unsigned int)timeb.millitm * 1000;
}
#define sockerr()	WSAGetLastError()
#ifndef EADDRINUSE
#define EADDRINUSE	WSAEADDRINUSE
#endif
#endif

#ifdef __OS2__
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <utils.h>
 
#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#define close(a)        soclose(a)
#endif

#include "qstat.h"

#ifndef FD_SETSIZE
#define FD_SETSIZE 64
#endif

/* Values set by command-line arguments
 */

int hostname_lookup= 0;		/* set if -H was specified */
int new_style= 1;		/* unset if -old was specified */
int n_retries= DEFAULT_RETRIES;
int retry_interval= DEFAULT_RETRY_INTERVAL;
int get_player_info= 0;
int get_server_rules= 0;
int up_servers_only= 0;
int no_full_servers= 0;
int no_empty_servers= 0;
int raw_display= 0;
char *raw_delimiter= "\t";
int player_address= 0;
int hex_player_names= 0;
int hexen2= 0;
int max_simultaneous= MAXFD_DEFAULT;
int raw_arg= 0;
int progress= 0;
int num_servers_total= 0;
int num_servers_returned= 0;
int num_servers_timed_out= 0;
int num_servers_down= 0;

#define DEFAULT_COLOR_NAMES_RAW		0
#define DEFAULT_COLOR_NAMES_DISPLAY	1
int color_names= -1;

#define SECONDS 0
#define CLOCK_TIME 1
#define STOPWATCH_TIME 2
#define DEFAULT_TIME_FMT_RAW		SECONDS
#define DEFAULT_TIME_FMT_DISPLAY	CLOCK_TIME
int time_format= -1;

struct qserver *servers= NULL;
int connected= 0;
time_t run_timeout= 0;
time_t start_time;

char *DOWN= "DOWN";
char *SYSERROR= "SYSERROR";
char *TIMEOUT= "TIMEOUT";
char *MASTER= "MASTER";
char *SERVERERROR= "ERROR";
char *DONE= "DONE";

unsigned int sample= 0x0f000000;
unsigned char *byteswap= (unsigned char *) &sample;
#define BIG_END (*byteswap)
#define LITTLE_END (!(*byteswap))

char *GET_QW_SERVERS= "GET_QW_SERVERS";
int display_prefix= 0;
char *current_filename;
int current_fileline;


/* MODIFY HERE
 * Change these functions to display however you want
 */
void
display_server(
    struct qserver *server
)
{
    char name[100], *prefix;

    if ( raw_display)  {
	raw_display_server( server);
	return;
    }

    if ( server->type == Q_SERVER)
	prefix= "QS  ";
    else if ( server->type == H2_SERVER)
	prefix= "H2S ";
    else if ( server->type == QW_SERVER)
	prefix= "QW  ";
    else if ( server->type == QW_MASTER)
	prefix= "QWM ";
    else if ( server->type == Q2_SERVER)
	prefix= "Q2  ";

    if ( server->server_name == DOWN)  {
	printf( "%s%-16s %10s\n", (display_prefix)?prefix:"",
		(hostname_lookup) ? server->host_name : server->arg, DOWN);
	return;
    }
    if ( server->server_name == TIMEOUT)  {
	if ( ! up_servers_only)
	    printf( "%s%-16s no response\n", (display_prefix)?prefix:"",
			(hostname_lookup) ? server->host_name : server->arg);
	return;
    }

    if ( server->type == QW_MASTER)  {
	display_qwmaster(server);
	return;
    }

    if ( no_full_servers && server->num_players >= server->max_players)
	return;

    if ( no_empty_servers && server->num_players == 0)
	return;

    if ( server->error != NULL)  {
	printf( "%s%-22s ERROR <%s>\n",
	    (display_prefix)?prefix:"",
	    (hostname_lookup) ? server->host_name : server->arg,
	    server->error);
	return;
    }

    if ( new_style)  {
	char *game= get_qw_game( server);
	printf( "%s%-22s %2d/%2d %8s %6d / %1d  %*s %s\n",
	    (display_prefix)?prefix:"",
	    (hostname_lookup) ? server->host_name : server->arg,
	    server->num_players, server->max_players,
	    (server->map_name) ? server->map_name : "?",
	    server->ping_total/server->n_requests,
	    server->n_retries,
	    (server->type == QW_SERVER || server->type == Q2_SERVER) ? 9 : 0, game,
	    (server->server_name) ? server->server_name : "?");
	if ( get_server_rules)
	    display_server_rules( server);
	if ( get_player_info)
	    display_player_info( server);
    }
    else  {
	sprintf( name, "\"%s\"", server->server_name);
	printf( "%-16s %10s map %s at %22s %d/%d players %d ms\n", 
	    (hostname_lookup) ? server->host_name : server->arg,
	    name, server->map_name,
	    server->address, server->num_players, server->max_players,
	    server->ping_total/server->n_requests);
    }
}

void
display_qwmaster( struct qserver *server)
{
    if ( server->error != NULL)
	printf( "QWM %-17s ERROR <%s>\n", 
		(hostname_lookup) ? server->host_name : server->arg,
		server->error);
    else
	printf( "QWM %-17s %d servers %6d / %1d\n",
	    (hostname_lookup) ? server->host_name : server->arg,
	    server->n_servers,
	    server->ping_total/server->n_requests,
	    server->n_retries);
}

void
display_header()
{
    printf( "%-16s %8s %8s %15s    %s\n", "ADDRESS", "PLAYERS", "MAP",
	"RESPONSE TIME", "NAME");
}

void
display_server_rules( struct qserver *server)
{
    struct rule *rule;
    int printed= 0;
    rule= server->rules;
    for ( ; rule != NULL; rule= rule->next)  {
	if ( (server->type != Q_SERVER && server->type != H2_SERVER) ||
		! is_default_rule( rule))  {
	    printf( "%c%s=%s", (printed)?',':'\t', rule->name, rule->value);
	    printed++;
	}
    }
    if ( printed)
	puts("");
}

void
display_player_info( struct qserver *server)
{
    if ( server->type == Q_SERVER || server->type == H2_SERVER)
	display_q_player_info(server);
    else if ( server->type == QW_SERVER)
	display_qw_player_info(server);
    else if ( server->type == Q2_SERVER)
	display_q2_player_info(server);
}

void
display_q_player_info( struct qserver *server)
{
    char fmt[128];
    struct player *player;

    strcpy( fmt, "\t#%-2d %3d frags %9s ");

    if ( color_names)
	strcat( fmt, "%9s:%-9s ");
    else
	strcat( fmt, "%2d:%-2d ");
    if ( player_address)
	strcat( fmt, "%22s ");
    else
	strcat( fmt, "%s");
    strcat( fmt, "%s\n");

    player= server->players;
    for ( ; player != NULL; player= player->next)  {
	printf( fmt,
		player->number,
		player->frags,
		play_time(player->connect_time,1),
		quake_color(player->shirt_color),
		quake_color(player->pants_color),
		(player_address)?player->address:"",
		escape( (unsigned char *) player->name));
    }
}

void
display_qw_player_info( struct qserver *server)
{
    char fmt[128];
    struct player *player;

    strcpy( fmt, "\t#%-6d %3d frags %6s@%-5s %8s");

    if ( color_names)
	strcat( fmt, "%9s:%-9s ");
    else
	strcat( fmt, "%2d:%-2d ");
    strcat( fmt, "%s\n");

    player= server->players;
    for ( ; player != NULL; player= player->next)  {
	printf( fmt,
		player->number,
		player->frags,
		play_time(player->connect_time,0),
		ping_time(player->ping),
		player->skin,
		quake_color(player->shirt_color),
		quake_color(player->pants_color),
		escape( (unsigned char *) player->name));
    }
}

void
display_q2_player_info( struct qserver *server)
{
    struct player *player;

    player= server->players;
    for ( ; player != NULL; player= player->next)  {
	printf( "\t%3d frags %8s  %s\n", 
		player->frags,
		ping_time(player->ping),
		escape( (unsigned char *) player->name));
    }
}

void
display_qw_user_info( struct qserver *server, struct qw_user *user)
{
    int i;
    char *color_fmt;

    if ( raw_display)  {
	raw_display_qw_user_info( server, user);
	return;
    }

    if ( user->error)  {
	printf( "%s\n", user->error);
	return;
    }

    printf( "%-16s [uid %6s] skl %5s %5s rnk %5s frg %5s dth %5s\n",
	user->name, user->uid,
        user->skill, user->efficiency, user->rank, user->frags, user->deaths
    );

    if ( color_names)
	color_fmt= "\t%-12s %s\n";
    else
	color_fmt= "\t%-12s %d\n";

    for ( i= 0; i < user->n_info; i++)
	if ( strcmp( user->info[i].key, "topcolor") == 0
		|| strcmp( user->info[i].key, "bottomcolor") == 0)
	    printf( color_fmt, user->info[i].key,
		quake_color( atoi(user->info[i].value)));
	else
	    printf( "\t%-12s %s\n", user->info[i].key, user->info[i].value);
}

void
display_qw_seen_info( struct qserver *server, char *seen)
{
    if ( raw_display)  {
	raw_display_qw_seen_info( server, seen);
	return;
    }

    printf( "%s\n", seen);
}

char *
get_qw_game( struct qserver *server)
{
    struct rule *rule;
    int printed= 0;
    rule= server->rules;
    for ( ; rule != NULL; rule= rule->next)
	if ( strcmp( rule->name, "*gamedir") == 0 ||
			strcmp( rule->name, "game") == 0)
	    return rule->value;
    return "";
}

/* Raw output for web master types
 */

#define RD raw_delimiter

void
raw_display_server( struct qserver *server)
{
    char *prefix;
    if ( server->type == Q_SERVER)
	prefix= "QS";
    else if ( server->type == H2_SERVER)
	prefix= "H2S";
    else if ( server->type == QW_SERVER)
	prefix= "QW";
    else if ( server->type == QW_MASTER)
	prefix= "QWM";
    else if ( server->type == Q2_SERVER)
	prefix= "Q2";

    if ( server->server_name == DOWN)  {
	if ( ! up_servers_only)
	    printf( "%s" "%.*s%.*s" "%s%s" "%s%s\n\n",
		prefix,
		raw_arg, RD, raw_arg, server->arg,
		RD, (hostname_lookup)?server->host_name:server->arg,
		RD, DOWN);
	return;
    }
    if ( server->server_name == TIMEOUT)  {
	if ( ! up_servers_only)
	    printf( "%s" "%.*s%.*s" "%s%s" "%s%s\n\n",
		prefix,
		raw_arg, RD, raw_arg, server->arg,
		RD, (hostname_lookup)?server->host_name:server->arg,
		RD, TIMEOUT);
	return;
    }

    if ( server->error != NULL)  {
        printf( "%s" "%.*s%.*s" "%s%s" "%s%s" "%s%s",
		prefix,
		raw_arg, RD, raw_arg, server->arg,
		RD, (hostname_lookup) ? server->host_name : server->arg,
		RD, "ERROR",
		RD, server->error);
    }
    else if ( server->type == Q_SERVER || server->type == H2_SERVER)  {
        printf( "%s" "%.*s%.*s" "%s%s" "%s%s" "%s%s" "%s%d" "%s%s" "%s%d" "%s%d" "%s%d" "%s%d",
		prefix,
		raw_arg, RD, raw_arg, server->arg,
		RD, (hostname_lookup) ? server->host_name : server->arg,
		RD, escape((unsigned char *) server->server_name),
		RD, server->address,
		RD, server->protocol_version,
		RD, server->map_name,
		RD, server->max_players,
		RD, server->num_players,
		RD, server->ping_total/server->n_requests,
		RD, server->n_retries
        );
    }
    else if ( server->type == QW_SERVER || server->type == Q2_SERVER)  {
        printf( "%s" "%.*s%.*s" "%s%s" "%s%s" "%s%s" "%s%d" "%s%d" "%s%d" "%s%d",
		prefix,
		raw_arg, RD, raw_arg, server->arg,
		RD, (hostname_lookup) ? server->host_name : server->arg,
		RD, escape((unsigned char *) server->server_name),
		RD, (server->map_name) ? server->map_name : "?",
		RD, server->max_players,
		RD, server->num_players,
		RD, server->ping_total/server->n_requests,
		RD, server->n_retries
        );
    }
    else if ( server->type == QW_MASTER)  {
        printf( "%s" "%.*s%.*s" "%s%s" "%s%d",
		prefix,
		raw_arg, RD, raw_arg, server->arg,
		RD, (hostname_lookup) ? server->host_name : server->arg,
		RD, server->n_servers
	);
    }
    puts("");

    if ( server->type == QW_MASTER || server->error != NULL)  {
	puts("");
	return;
    }

    if ( get_server_rules)
	raw_display_server_rules( server);
    if ( get_player_info)
	raw_display_player_info( server);
    puts("");
}

void
raw_display_server_rules( struct qserver *server)
{
    struct rule *rule;
    int printed= 0;
    rule= server->rules;
    for ( ; rule != NULL; rule= rule->next)  {
	printf( "%s%s=%s", (printed)?RD:"", rule->name, rule->value);
	printed++;
    }
    if ( server->missing_rules)
	printf( "%s?", (printed)?RD:"");
    puts("");
}

void
raw_display_player_info( struct qserver *server)
{
    if ( server->type == Q_SERVER || server->type == H2_SERVER)
	raw_display_q_player_info( server);
    else if ( server->type == QW_SERVER)
	raw_display_qw_player_info( server);
    else if ( server->type == Q2_SERVER)
	raw_display_q2_player_info( server);
}

void
raw_display_q_player_info( struct qserver *server)
{
    char fmt[128];
    struct player *player;

    strcpy( fmt, "%d" "%s%s" "%s%s" "%s%d" "%s%s");
    if ( color_names)
	strcat( fmt, "%s%s" "%s%s");
    else
	strcat( fmt, "%s%d" "%s%d");

    player= server->players;
    for ( ; player != NULL; player= player->next)  {
	printf( fmt,
		player->number,
		RD, escape( (unsigned char *) player->name),
		RD, player->address,
		RD, player->frags,
		RD, play_time(player->connect_time,1),
		RD, quake_color(player->shirt_color),
		RD, quake_color(player->pants_color)
	);
	puts("");
    }
}

void
raw_display_qw_player_info( struct qserver *server)
{
    char fmt[128];
    struct player *player;

    strcpy( fmt, "%d" "%s%s" "%s%d" "%s%s");
    if ( color_names)
	strcat( fmt, "%s%s" "%s%s");
    else
	strcat( fmt, "%s%d" "%s%d");
    strcat( fmt, "%s%d" "%s%s");

    player= server->players;
    for ( ; player != NULL; player= player->next)  {
	printf( fmt,
		player->number,
		RD, escape( (unsigned char *) player->name),
		RD, player->frags,
		RD, play_time(player->connect_time,1),
		RD, quake_color(player->shirt_color),
		RD, quake_color(player->pants_color),
		RD, player->ping,
		RD, player->skin
	);
	puts("");
    }
}

void
raw_display_q2_player_info( struct qserver *server)
{
    char fmt[128];
    struct player *player;

    strcpy( fmt, "%s" "%s%d" "%s%d");

    player= server->players;
    for ( ; player != NULL; player= player->next)  {
	printf( fmt,
		escape( (unsigned char *) player->name),
		RD, player->frags,
		RD, player->ping
	);
	puts("");
    }
}

void
raw_display_qw_user_info( struct qserver *server, struct qw_user *user)
{
    int i;
    char *color_fmt;

    if ( user->error)  {
	printf( "QWU%s%s\n", RD, user->error);
	return;
    }

    printf( "%s" "%s%s%s%s" "%s%s%s%s" "%s%s%s%s" "%s%s%s%s" "%s%s%s%s"
		"%s%s%s%s" "%s%s%s%s",
	"QWU",
	RD, "name", RD, user->name,
	RD, "userid", RD, user->uid,
        RD, "skill", RD, user->skill,
	RD, "efficiency", RD, user->efficiency,
	RD, "rank", RD, user->rank,
	RD, "frags", RD, user->frags,
	RD, "deaths", RD, user->deaths
    );

    if ( color_names)
	color_fmt= "%s%s%s%s";
    else
	color_fmt= "%s%s%s%d";

    for ( i= 0; i < user->n_info; i++)
	if ( strcmp( user->info[i].key, "topcolor") == 0
		|| strcmp( user->info[i].key, "bottomcolor") == 0)
	    printf( color_fmt, RD, user->info[i].key,
		RD, quake_color( atoi(user->info[i].value)));
	else
	    printf( "%s%s%s%s",
		RD, user->info[i].key, RD, user->info[i].value);
    puts("");
    puts("");
}

void
raw_display_qw_seen_info( struct qserver *server, char *seen)
{
    printf( "QWE%s%s\n\n", RD, seen);
}

void
display_progress()
{
    fprintf( stderr, "\r%d/%d (%d timed out, %d down)",
	num_servers_returned+num_servers_timed_out+num_servers_down,
	num_servers_total,
	num_servers_timed_out,
	num_servers_down);
}

/* ----- END MODIFICATION ----- Don't need to change anything below here. */


/* These packets are fixed size, so don't bother rebuilding them each time.
 * This only works because the packets are an even number in size.
 */
struct {
    unsigned char flag1;
    unsigned char flag2;
    unsigned short length;
    unsigned char op_code;
    char name[6];
    unsigned char version;
} qserverinfo =
	{ Q_FLAG1, Q_FLAG2, sizeof(qserverinfo), Q_CCREQ_SERVER_INFO,
	"QUAKE", Q_NET_PROTOCOL_VERSION };

struct {
    unsigned char flag1;
    unsigned char flag2;
    unsigned short length;
    unsigned char op_code;
    char name[8];
    unsigned char version;
} h2serverinfo =
	{ Q_FLAG1, Q_FLAG2, sizeof(h2serverinfo), Q_CCREQ_SERVER_INFO,
	"HEXENII", HEXEN2_NET_PROTOCOL_VERSION };

struct {
    unsigned char flag1;
    unsigned char flag2;
    unsigned short length;
    unsigned char op_code;
    unsigned char player_number;
} qplayerinfo =
	{ Q_FLAG1, Q_FLAG2, sizeof(qplayerinfo), Q_CCREQ_PLAYER_INFO, 0 };

struct {
    char prefix[4];
    char command[16];
} qwserverstatus = { QW_REQUEST_PREFIX, 's', 't', 'a', 't', 'u', 's', '\n',
	0, 0, 0, 0, 0, 0, 0};

/* Misc flags
 */

char * NO_SERVER_RULES= NULL;
int NO_PLAYER_INFO= 0xffff;
struct timeval packet_recv_time;
int server_types= 0;
#define FORCE 1

/* Print an error message and the program usage notes
 */

void
usage( char *msg, char **argv)
{
    if ( msg)
	fprintf( stderr, msg);

    printf( "Usage: %s [options ...] [-retry retries] [-interval interval]\n",
	argv[0]);
    printf( "\t[-raw delimiter] [-qw host[:port]] [-qws host[:port]]\n");
    printf( "\t[-h2s host[:port]] [-f file] host[:port] ...\n");
    printf( "Where host is an IP address or host name\n");
    printf( "Port defaults to %d if not specified\n", DEFAULT_PORT);
    printf( "-H\t\tresolve host names\n");
    printf( "-R\t\tserver rules\n");
    printf( "-P\t\tplayer info\n");
    printf( "-u\t\tonly display servers that are up\n");
    printf( "-nf\t\tdo not display full servers\n");
    printf( "-ne\t\tdo not display empty servers\n");
    printf( "-cn\t\tdisplay color names instead of numbers\n");
    printf( "-ncn\t\tdisplay color numbers instead of names\n");
    printf( "-tc\t\tdisplay time in clock format (DhDDmDDs)\n");
    printf( "-tsw\t\tdisplay time in stop-watch format (DD:DD:DD)\n");
    printf( "-ts\t\tdisplay time in seconds\n");
    printf( "-pa\t\tdisplay player address\n");
    printf( "-hpn\t\tdisplay player names in hex\n");
    printf( "-old\t\told style display\n");
    printf( "-progress\tdisplay progress meter (text only)\n");
    printf( "-retry\t\tnumber of retries, default is %d\n", DEFAULT_RETRIES);
    printf( "-interval\tinterval between retries, default is %.2lf seconds\n",
	DEFAULT_RETRY_INTERVAL / 1000.0);
    printf( "-timeout\ttotal time in seconds before giving up\n");
    printf( "-maxsimultaneous\tset max simultaneous queries\n");
    printf( "-raw\t\toutput in raw format using delimiter\n");
    printf( "-f\t\tread hosts from file\n");
    printf( "-qw\t\tget server list from QuakeWorld master server\n");
    printf( "-qws\t\tdisplay stats for QuakeWorld server\n");
    printf( "-h2s\t\tdisplay stats for Hexen II server\n");
    printf( "-hexen2\t\tquery using Hexen II protocol\n");
    printf( "qstat version %s\n", VERSION);
    exit(0);
}

main( int argc, char *argv[])
{
    int pktlen, rc, maxfd;
    long pkt_data[PACKET_LEN/sizeof(long)];
    char *pkt= (char*)&pkt_data[0];
    fd_set read_fds;
    struct timeval timeout;
    int arg, timeout_milli, n_files, i;
    int n_qw_masters, n_qw_servers, n_h2_servers;
    struct qserver *server;
    char **files, **qw_masters, **qw_servers, **h2_servers;
 
#ifdef _WIN32
    WORD version= MAKEWORD(1,1);
    WSADATA wsa_data;
    if ( WSAStartup(version,&wsa_data) != 0)  {
	fprintf( stderr, "Could not open winsock\n");
	exit(1);
    }
#endif

    if ( argc == 1)
	usage(NULL,argv);

    files= (char **) malloc( sizeof(char*) * (argc/2));
    n_files= 0;
    qw_masters= (char **) malloc( sizeof(char*) * (argc/2));
    qw_servers= (char **) calloc( 1, sizeof(char*) * (argc/2));
    h2_servers= (char **) calloc( 1, sizeof(char*) * (argc/2));
    n_qw_masters= 0;
    n_qw_servers= 0;
    n_h2_servers= 0;

    for ( arg= 1; arg < argc; arg++)  {
	if ( argv[arg][0] != '-')
	    break;
	if ( strcmp( argv[arg], "-f") == 0)  {
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -f\n", argv);
	    files[n_files++]= argv[arg];
	}
	else if ( strcmp( argv[arg], "-retry") == 0)  {
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -retry\n", argv);
	    n_retries= atoi( argv[arg]);
	    if ( n_retries <= 0)  {
		fprintf( stderr, "retries must be greater than zero\n");
		exit(1);
	    }
	}
	else if ( strcmp( argv[arg], "-interval") == 0)  {
	    double value= 0.0;
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -interval\n", argv);
	    sscanf( argv[arg], "%lf", &value);
	    if ( value < 0.1)  {
		fprintf( stderr, "retry interval must be greater than 0.1\n");
		exit(1);
	    }
	    retry_interval= (int)(value * 1000);
	}
	else if ( strcmp( argv[arg], "-H") == 0)
	    hostname_lookup= 1;
	else if ( strcmp( argv[arg], "-u") == 0)
	    up_servers_only= 1;
	else if ( strcmp( argv[arg], "-nf") == 0)
	    no_full_servers= 1;
	else if ( strcmp( argv[arg], "-ne") == 0)
	    no_empty_servers= 1;
	else if ( strcmp( argv[arg], "-old") == 0)
	    new_style= 0;
	else if ( strcmp( argv[arg], "-P") == 0)
	    get_player_info= 1;
	else if ( strcmp( argv[arg], "-R") == 0)
	    get_server_rules= 1;
	else if ( strcmp( argv[arg], "-raw") == 0)  {
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -raw\n", argv);
	    raw_delimiter= argv[arg];
	    raw_display= 1;
	}
	else if ( strcmp( argv[arg], "-ncn") == 0)  {
	    color_names= 0;
 	}
	else if ( strcmp( argv[arg], "-cn") == 0)  {
	    color_names= 1;
 	}
	else if ( strcmp( argv[arg], "-tc") == 0)  {
	    time_format= CLOCK_TIME;
	}
	else if ( strcmp( argv[arg], "-tsw") == 0)  {
	    time_format= STOPWATCH_TIME;
	}
	else if ( strcmp( argv[arg], "-ts") == 0)  {
	    time_format= SECONDS;
	}
	else if ( strcmp( argv[arg], "-pa") == 0)  {
	    player_address= 1;
	}
	else if ( strcmp( argv[arg], "-hpn") == 0)  {
	    hex_player_names= 1;
	}
	else if ( strcmp( argv[arg], "-qw") == 0)  {
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -qw\n", argv);
	    qw_masters[n_qw_masters++]= argv[arg];
	}
	else if ( strcmp( argv[arg], "-qws") == 0)  {
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -qws\n", argv);
	    qw_servers[n_qw_servers++]= argv[arg];
	}
	else if ( strcmp( argv[arg], "-hexen2") == 0)  {
	    hexen2= 1;
 	}
	else if ( strcmp( argv[arg], "-maxsimultaneous") == 0)  {
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -maxsimultaneous\n", argv);
	    max_simultaneous= atoi(argv[arg]);
	    if ( max_simultaneous <= 0)
		usage( "value for -maxsimultaneous must be > 0\n", argv);
	    if ( max_simultaneous > FD_SETSIZE)
		max_simultaneous= FD_SETSIZE;
 	}
	else if ( strcmp( argv[arg], "-h2s") == 0)  {
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -h2s\n", argv);
	    h2_servers[n_h2_servers++]= argv[arg];
	}
	else if ( strcmp( argv[arg], "-raw-arg") == 0)  {
	    raw_arg= 1000;
 	}
	else if ( strcmp( argv[arg], "-timeout") == 0)  {
	    arg++;
	    if ( arg >= argc)
		usage( "missing argument for -timeout\n", argv);
	    run_timeout= atoi( argv[arg]);
	    if ( run_timeout <= 0)
		usage( "value for -timeout must be > 0\n", argv);
	}
	else if ( strcmp( argv[arg], "-progress") == 0)  {
	    progress= 1;
 	}
	else  {
	    fprintf( stderr, "unknown option \"%s\"\n", argv[arg]);
	    usage(NULL,argv);
	}
    }

    start_time= time(0);

    for ( i= 0; i < n_files; i++)
	add_file( files[i]);
    for ( ; arg < argc; arg++)
	add_qserver( argv[arg], (hexen2)?H2_SERVER:Q_SERVER);
    for ( i= 0; i < n_h2_servers; i++)
	add_qserver( h2_servers[i], H2_SERVER);
    for ( i= 0; i < n_qw_servers; i++)
	add_qserver( qw_servers[i], QW_SERVER);
    for ( i= 0; i < n_qw_masters; i++)
	add_qserver( qw_masters[i], QW_MASTER);

    if ( servers == NULL)
	exit(1);

    if ( color_names == -1)
	color_names= ( raw_display) ? DEFAULT_COLOR_NAMES_RAW :
		DEFAULT_COLOR_NAMES_DISPLAY;

    if ( time_format == -1)
	time_format= ( raw_display) ? DEFAULT_TIME_FMT_RAW :
		DEFAULT_TIME_FMT_DISPLAY;

    if ( (server_types & QW_MASTER) || count_bits(server_types) > 1)
	display_prefix= 1;

    if ( new_style && ! raw_display)
	display_header();

    qserverinfo.length= htons( qserverinfo.length);
    h2serverinfo.length= htons( h2serverinfo.length);
    qplayerinfo.length= htons( qplayerinfo.length);

    bind_sockets();

    while (connected)  {
	FD_ZERO( &read_fds);
	maxfd= set_fds( &read_fds);

	if ( progress)
	    display_progress();
	get_next_timeout( &timeout);
	rc= select( maxfd+1, &read_fds, NULL, NULL, &timeout);

	if (rc == 0)  {
	    if ( run_timeout && time(0)-start_time >= run_timeout)
		return 0;
	    send_packets();
	    bind_sockets();
	    continue;
	}
	if (rc == SOCKET_ERROR)  {
	    perror("select");
	    return -1;
	}

	gettimeofday( &packet_recv_time, NULL);
	for ( server= servers; server != NULL; server= server->next)  {
	    if ( server->fd == -1)
	        continue;
	    if ( ! FD_ISSET( server->fd, &read_fds))
		continue;
	    if ((pktlen= recv( server->fd, pkt, sizeof(pkt_data), 0)) ==
			SOCKET_ERROR)  {
		if ( connection_refused())  {
		    server->server_name= DOWN;
		    close(server->fd);
		    server->fd= -1;
		    num_servers_down++;
		    if ( ! up_servers_only)
			display_server( server);
		    connected--;
		}
		continue;
	    }

	    deal_with_packet( server, pkt, pktlen);
	}
	if ( run_timeout && time(0)-start_time >= run_timeout)
	    return 0;
	if ( ! connected)
	    bind_sockets();
    }

    return 0;
}


void
add_file( char *filename)
{
    FILE *file;
    char name[200];
    int type;

    if ( strcmp( filename, "-") == 0)  {
	file= stdin;
	current_filename= NULL;
    }
    else  {
	file= fopen( filename, "r");
	current_filename= filename;
    }
    current_fileline= 1;

    if ( file == NULL)  {
	perror( filename);
	return;
    }
    for ( ; fscanf( file, "%s", name) == 1; current_fileline++)  {
	if ( strcmp( name, "QS") == 0)
	    type= Q_SERVER;
	else if ( strcmp( name, "H2S") == 0)
	    type= H2_SERVER;
	else if ( strcmp( name, "QW") == 0)
	    type= QW_SERVER;
	else if ( strcmp( name, "Q2") == 0)
	    type= Q2_SERVER;
	else if ( strcmp( name, "QWM") == 0)
	    type= QW_MASTER;
	else  {
	    add_qserver( name, (hexen2)?H2_SERVER:Q_SERVER);
	    continue;
	}
	if ( fscanf( file, "%s", name) == 1)
	    add_qserver( name, type);
    }

    if ( file != stdin)
	fclose(file);

    current_fileline= 0;
}

void
print_file_location()
{
    if ( current_fileline != 0)
	fprintf( stderr, "%s:%d: ", current_filename?current_filename:"<stdin>",
		current_fileline);
}

int
add_qserver( char *arg, int type)
{
    struct sockaddr_in addr;
    struct hostent *ent= NULL;
    struct hostent temp_ent;
    struct hostent *name_ent= NULL;
    struct qserver *server;
    int i, port, nf;
    char **a, *colon, *arg_copy, *s;
    char *h_addr_list[MAXIP];
    char *uid= NULL, *password= NULL;
    unsigned long ipaddr;
 
    if ( run_timeout && time(0)-start_time >= run_timeout)
	exit(0);

    if ( type == Q_SERVER)
	port= DEFAULT_PORT;
    else if ( type == H2_SERVER)
	port= HEXEN2_DEFAULT_PORT;
    else if ( type == QW_SERVER)
	port= QW_DEFAULT_PORT;
    else if ( type == QW_MASTER)
	port= QW_MASTER_DEFAULT_PORT;
    else if ( type == Q2_SERVER)
	port= QUAKE2_DEFAULT_PORT;
    else
	port= DEFAULT_PORT;

    arg_copy= strdup(arg);

    colon= arg;
    for ( nf= 1; *colon && (colon= strchr( colon, ':')) != NULL; nf++)
	*colon++= '\0';

    s= arg + strlen(arg)+1;
    if ( nf > 1)  {
	sscanf( s, "%d", &port);
	s+= strlen(s)+1;
	if ( nf > 2)  {
	    arg_copy[s-arg-1]= '\0';
	    uid= strdup(s);
	    s+= strlen(s)+1;
	    if ( nf > 3)  {
		password= strdup(s);
		s+= strlen(s)+1;
	    }
	}
    }
    if ( arg_copy[strlen(arg_copy)-1] == ':')
	arg_copy[strlen(arg_copy)-1]= '\0';

    ipaddr= inet_addr(arg);
    if ( ipaddr == INADDR_NONE)
	ent= gethostbyname(arg);

    if ( hostname_lookup && ipaddr != INADDR_NONE)
	name_ent= gethostbyaddr( (char*)&ipaddr, sizeof(ipaddr), AF_INET);

    if ( ent == NULL && ipaddr != INADDR_NONE)  {
	/* Maybe gethostbyname() doesn't parse dotted IP addresses,
	 * try building a hostent by hand.
	 */
	if ( ipaddr == INADDR_NONE)  {
	     print_file_location();
	     fprintf( stderr, "%s: %s\n", arg, strherror(h_errno));
	     return -1;
	}
	ent= &temp_ent;
	ent->h_name= arg;
	ent->h_aliases= NULL;
	ent->h_addrtype= 2;
	ent->h_length= 4;
	h_addr_list[0]= (char *) &ipaddr;
	h_addr_list[1]= NULL;
	ent->h_addr_list= h_addr_list;
    }
    else if ( ent == NULL)  {
	print_file_location();
        fprintf( stderr, "%s: %s\n", arg, strherror(h_errno));
        return -1;
    }

    server= (struct qserver *) calloc( 1, sizeof( struct qserver));
    server->arg= arg_copy;
    server->host_name= strdup((name_ent)?name_ent->h_name:ent->h_name);
    for ( a= ent->h_addr_list, i= 0; *a != NULL && i < MAXIP; a++, i++)  {
	memcpy( &server->ip[i].s_addr, *a, sizeof(server->ip[i].s_addr));
    }
    server->port= port;
    server->type= type;
    init_qserver( server);

    server->next= servers;
    servers= server;

    server_types|= type;

    return 0;
}

int
add_qserver_byaddr( unsigned long ipaddr, unsigned short port, int type)
{
    char arg[36];
    struct qserver *server;
    struct hostent *name_ent= NULL;

    if ( run_timeout && time(0)-start_time >= run_timeout)
	exit(0);

    server= (struct qserver *) calloc( 1, sizeof( struct qserver));
    sprintf( arg, "%d.%d.%d.%d:%hu", ipaddr>>24, (ipaddr>>16)&0xff,
	(ipaddr>>8)&0xff, ipaddr&0xff, port);
    server->arg= strdup(arg);

    if ( hostname_lookup)
	name_ent= gethostbyaddr( (char*)&ipaddr, sizeof(ipaddr), AF_INET);
    server->host_name= strdup(name_ent ? name_ent->h_name : arg);

    server->ip[0].s_addr= htonl( ipaddr);
    server->port= port;
    server->type= type;
    init_qserver( server);

    server->next= servers;
    servers= server;
    return 0;
}

void
init_qserver( struct qserver *server)
{
    server->server_name= NULL;
    server->map_name= NULL;
    server->num_players= 0;
    server->fd= -1;
    server->retry1= n_retries;
    server->retry2= n_retries;
    server->n_retries= 0;
    server->ping_total= 0;
    server->n_packets= 0;
    server->n_requests= 0;

    server->n_servers= 0;
    server->error= NULL;

    server->next_rule= (get_server_rules) ? "" : NO_SERVER_RULES;
    server->next_player_info= (get_player_info) ? 0 : NO_PLAYER_INFO;

    server->n_player_info= 0;
    server->players= NULL;
    server->n_rules= 0;
    server->rules= NULL;
    server->missing_rules= 0;

    num_servers_total++;
}

/* Functions for binding sockets to Quake servers
 */
int
bind_qserver( struct qserver *server)
{
    struct sockaddr_in addr;

    if ((server->fd= socket( AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET) {
	if ( sockerr() == EMFILE)  {
	    server->fd= -1;
	    return -1;
	}
        perror( "socket" );
	server->server_name= SYSERROR;
        return -1;
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(0);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    memset( &(addr.sin_zero), 0, sizeof(addr.sin_zero) );

    if ( bind( server->fd, (struct sockaddr *)&addr,
		sizeof(struct sockaddr)) == SOCKET_ERROR) {
	if ( sockerr() != EADDRINUSE)  {
	    perror( "bind" );
	    server->server_name= SYSERROR;
	}
	close(server->fd);
	server->fd= -1;
        return -1;
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(server->port);
    addr.sin_addr = server->ip[0];
    memset( &(addr.sin_zero), 0, sizeof(addr.sin_zero) );

    if ( connect( server->fd, (struct sockaddr *)&addr, sizeof(addr)) ==
		SOCKET_ERROR)  {
	perror( "connect");
	server->server_name= SYSERROR;
	close(server->fd);
	server->fd= -1;
	return -1;
    }
    return 0;
}

void
bind_sockets()
{
    struct qserver *server= servers;

    while ( server != NULL)  {
	if ( server->server_name == NULL && server->fd == -1 &&
		connected < max_simultaneous)  {
	    if ( bind_qserver( server) == 0)  {
		send_server_request_packet( server);
		connected++;
	    }
	}
	server= server->next;
    }
}


/* Functions for sending packets
 */
void
send_packets()
{
    struct qserver *server= servers;
    struct timeval now;
    int rc;

    gettimeofday( &now, NULL);

    for ( ; server != NULL; server= server->next)  {
	if ( server->fd == -1)
	    continue;
	if ( server->server_name == NULL)  {
	    if ( server->retry1 != n_retries &&
		    time_delta( &now, &server->packet_time1) <
			(retry_interval*(n_retries-server->retry1+1)))
		continue;
	    if ( ! server->retry1)  {
		cleanup_qserver( server, 0);
		continue;
	    }
	    send_server_request_packet( server);
	    continue;
	}
	if ( server->next_rule != NO_SERVER_RULES)  {
	    if ( server->retry1 != n_retries &&
		    time_delta( &now, &server->packet_time1) <
			(retry_interval*(n_retries-server->retry1+1)))
		continue;
	    if ( ! server->retry1)  {
		server->next_rule= NULL;
		server->missing_rules= 1;
		cleanup_qserver( server, 0);
		continue;
	    }
	    send_rule_request_packet( server);
	}
	if ( server->next_player_info < server->num_players)  {
	    if ( server->retry2 != n_retries &&
		    time_delta( &now, &server->packet_time2) <
			(retry_interval*(n_retries-server->retry2+1)))
		continue;
	    if ( ! server->retry2)  {
		server->next_player_info++;
		if ( server->next_player_info >= server->num_players)  {
		    cleanup_qserver( server, 0);
		    continue;
		}
		server->retry2= n_retries;
	    }
	    send_player_request_packet( server);
	}
    }
}

/* Send the first packet to a server
 */
int
send_server_request_packet( struct qserver *server)
{
    if ( server->type == Q_SERVER || server->type == H2_SERVER)
	return send_qserver_request_packet( server);
    else if ( server->type == QW_SERVER || server->type == Q2_SERVER)
	return send_qwserver_request_packet( server);
    else if ( server->type == QW_MASTER)
	return send_qwmaster_request_packet( server);
    else
	return -1;
}

/* First packet for a normal Quake server
 */
int
send_qserver_request_packet( struct qserver *server)
{
    int rc;
    if ( server->type == H2_SERVER)
	rc= send( server->fd, (const char *)&h2serverinfo,
		sizeof(h2serverinfo), 0);
    else
	rc= send( server->fd, (const char *)&qserverinfo,
		sizeof(qserverinfo), 0);
    if ( rc == SOCKET_ERROR)
	perror( "send");
    if ( server->retry1 == n_retries)  {
	gettimeofday( &server->packet_time1, NULL);
	server->n_requests++;
    }
    else
	server->n_retries++;
    server->retry1--;
    server->n_packets++;
    return 0;
}

/* First packet for a QuakeWorld server
 */
int
send_qwserver_request_packet( struct qserver *server)
{
    int rc;

    rc= send( server->fd, (const char *)&qwserverstatus,
	sizeof(qwserverstatus), 0);
    if ( rc == SOCKET_ERROR)
	perror( "send");
    if ( server->retry1 == n_retries)  {
	gettimeofday( &server->packet_time1, NULL);
	server->n_requests++;
    }
    else
	server->n_retries++;
    server->retry1--;
    server->n_packets++;
    return 0;
}

/* First packet for a QuakeWorld master server
 */
int
send_qwmaster_request_packet( struct qserver *server)
{
    int rc;
    char pkt[QW_REQUEST_LENGTH];

    memset( pkt, 0, sizeof(pkt));
    sprintf( pkt, "%c%c", QW_GET_SERVERS, QW_NEWLINE);

    rc= send( server->fd, pkt, sizeof(pkt), 0);
    if ( rc == SOCKET_ERROR)
	perror( "send");
    if ( server->retry1 == n_retries)  {
	gettimeofday( &server->packet_time1, NULL);
	server->n_requests++;
    }
    else
	server->n_retries++;
    server->retry1--;
    server->n_packets++;
    return 0;
}

int
send_rule_request_packet( struct qserver *server)
{
    struct qpacket request= {Q_FLAG1,Q_FLAG2,0,Q_CCREQ_RULE_INFO,""};
    int rc, len;

    strcpy( (char*)request.data, server->next_rule);
    len= Q_HEADER_SIZE + strlen((char*)request.data) + 1;
    request.length= htons( (short)len);
    rc= send( server->fd, (const char *)&request, len, 0);
    if ( rc == SOCKET_ERROR)
	perror( "send");
    if ( server->retry1 == n_retries)  {
	gettimeofday( &server->packet_time1, NULL);
	server->n_requests++;
    }
    else
	server->n_retries++;
    server->retry1--;
    server->n_packets++;
    return 0;
}

int
send_player_request_packet( struct qserver *server)
{
    int rc;

    qplayerinfo.player_number= server->next_player_info;
    rc= send( server->fd, (const char *)&qplayerinfo, sizeof(qplayerinfo), 0);
    if ( rc == SOCKET_ERROR)
	perror( "send");
    if ( server->retry2 == n_retries)  {
	gettimeofday( &server->packet_time2, NULL);
	server->n_requests++;
    }
    else
	server->n_retries++;
    server->retry2--;
    server->n_packets++;
    return 0;
}

/* Functions for figuring timeouts and when to give up
 */
void
cleanup_qserver( struct qserver *server, int force)
{
    int close_it= force;
    if ( server->server_name == NULL)  {
	server->server_name= TIMEOUT;
	close_it= 1;
	num_servers_timed_out++;
    }
    else if ( server->type == QW_SERVER || server->type == QW_MASTER ||
		server->type == Q2_SERVER)
	close_it= 1;
    else if ( server->next_rule == NO_SERVER_RULES &&
		server->next_player_info >= server->num_players)
	close_it= 1;

    if ( close_it)  {
	close( server->fd);
	server->fd= -1;
	connected--;
	if ( server->server_name != TIMEOUT)
	    num_servers_returned++;
	display_server( server);
    }
}

void
get_next_timeout( struct timeval *timeout)
{
    struct qserver *server= servers;
    struct timeval now;
    int diff1, diff2, diff, smallest= retry_interval;
    gettimeofday( &now, NULL);
    for ( ; server != NULL; server= server->next)  {
	if ( server->fd == -1)
	    continue;
	diff2= 0xffff;
	diff1= 0xffff;
	if ( server->server_name == NULL)
	    diff1= retry_interval*(n_retries-server->retry1+1) -
		time_delta( &now, &server->packet_time1);
	else  {
	    if ( server->next_rule != NULL)
		diff1= retry_interval*(n_retries-server->retry1+1) -
			time_delta( &now, &server->packet_time1);
	    if ( server->next_player_info < server->num_players)
		diff2= retry_interval*(n_retries-server->retry2+1) -
			time_delta( &now, &server->packet_time2);
	}
	diff= (diff1<diff2)?diff1:diff2;
	if ( diff < smallest)
	    smallest= diff;
    }
    if ( smallest < 10)
	smallest= 10;
    timeout->tv_sec= smallest / 1000;
    timeout->tv_usec= (smallest % 1000) * 1000;
}

int
set_fds( fd_set *fds)
{
    struct qserver *server= servers;
    int maxfd= 1;
    while ( server != NULL)  {
	if ( server->fd != -1)  {
	    FD_SET( server->fd, fds);
	    if ( server->fd > maxfd)
		maxfd= server->fd;
	}
	server= server->next;
    }
    return maxfd;
}


/* Functions for handling response packets
 */
void
deal_with_packet( struct qserver *server, char *rawpkt, int pktlen)
{
    if ( server->type == Q_SERVER || server->type == H2_SERVER)
 	deal_with_q_packet( server, rawpkt, pktlen);
    else if ( server->type == QW_SERVER || server->type == Q2_SERVER)
 	deal_with_qw_packet( server, rawpkt, pktlen);
    else if ( server->type == QW_MASTER)
 	deal_with_qwmaster_packet( server, rawpkt, pktlen);
}

/* Packet from normal Quake server
 */
void
deal_with_q_packet( struct qserver *server, char *rawpkt, int pktlen)
{
    struct qpacket *pkt= (struct qpacket *)rawpkt;
    int rc;

    if ( ntohs( pkt->length) != pktlen)  {
	fprintf( stderr, "%s Ignoring bogus packet; length %d != %d\n",
		server->arg, ntohs( pkt->length), pktlen);
	cleanup_qserver(server,FORCE);
	return;
    }

    rawpkt[pktlen]= '\0';

    switch ( pkt->op_code)  {
    case Q_CCREP_ACCEPT:
    case Q_CCREP_REJECT:
	return;
    case Q_CCREP_SERVER_INFO:
	server->ping_total+= time_delta( &packet_recv_time,
		&server->packet_time1);
	rc= server_info_packet( server, pkt, pktlen-Q_HEADER_SIZE);
	break;
    case Q_CCREP_PLAYER_INFO:
	server->ping_total+= time_delta( &packet_recv_time,
		&server->packet_time2);
	rc= player_info_packet( server, pkt, pktlen-Q_HEADER_SIZE);
	break;
    case Q_CCREP_RULE_INFO:
	server->ping_total+= time_delta( &packet_recv_time,
		&server->packet_time1);
	rc= rule_info_packet( server, pkt, pktlen-Q_HEADER_SIZE);
	break;
    case Q_CCREQ_CONNECT:
    case Q_CCREQ_SERVER_INFO:
    case Q_CCREQ_PLAYER_INFO:
    case Q_CCREQ_RULE_INFO:
    default:
	return;
    }

    if ( rc == -1)
	fprintf( stderr, "%s error on packet opcode %x\n", server->arg,
		(int)pkt->op_code);

    cleanup_qserver( server, (rc == -1) ? FORCE : 0);
}

/* Packet from QuakeWorld server
 */
void
deal_with_qw_packet( struct qserver *server, char *rawpkt, int pktlen)
{
    char *pkt= &rawpkt[5];
    char *key, *value, *end;
    int len, rc, odd= 0, complete= 0;

    server->ping_total+= time_delta( &packet_recv_time,
		&server->packet_time1);

    if ( rawpkt[0] != '\377' || rawpkt[1] != '\377' ||
	    rawpkt[2] != '\377' || rawpkt[3] != '\377')  {
	fprintf( stderr, "Odd packet from QW server, processing anyway ...\n");
	print_packet( rawpkt, pktlen);
	odd= 1;
    }
    if ( rawpkt[4] == 'n')  {
	server->type= QW_SERVER;
	deal_with_q1qw_packet( server, rawpkt, pktlen);
    }
    else if ( strncmp( &rawpkt[4], "print\n\\", 7) == 0)  {
	server->type= Q2_SERVER;
	deal_with_q2_packet( server, rawpkt, pktlen);
    }
    else  {
	fprintf( stderr, "Odd packet from QW server, skipping ...\n");
	print_packet( rawpkt, pktlen);
	cleanup_qserver( server, 0);
	return;
    }
}

void
deal_with_q1qw_packet( struct qserver *server, char *rawpkt, int pktlen)
{
    char *key, *value, *end;
    struct rule *rule;
    struct player *player;
    int len, rc, odd= 0, complete= 0;
    char *pkt= &rawpkt[5];

    while ( *pkt && pkt-rawpkt < pktlen)  {
	if ( *pkt == '\\')  {
	    pkt++;
	    end= strchr( pkt, '\\');
	    if ( end == NULL)
		break;
	    *end= '\0';
	    key= pkt;
	    pkt+= strlen(pkt)+1;
	    end= strchr( pkt, '\\');
	    if ( end == NULL)
		end= strchr( pkt, '\n');
	    value= (char*) malloc(end-pkt+1);
	    memcpy( value, pkt, end-pkt);
	    value[end-pkt]= '\0';
	    pkt= end;
	    if ( strcmp( key, "hostname") == 0)
		server->server_name= value;
	    else if  ( strcmp( key, "map") == 0)
		server->map_name= value;
	    else if  ( strcmp( key, "maxclients") == 0)
		server->max_players= atoi(value);
	    else  {
		rule= (struct rule *) malloc( sizeof( struct rule));
		rule->name= strdup(key);
		rule->value= value;
		rule->next= server->rules;
		server->rules= rule;
		server->n_rules++;
	    }
	}
	else if ( *pkt == '\n')  {
	    pkt++;
	    if ( *pkt == '\0')
		break;
	    player= (struct player *) calloc( 1, sizeof( struct player));
	    rc= sscanf( pkt, "%d %d %d %d %n", &player->number, &player->frags,
		&player->connect_time, &player->ping, &len);
	    if ( rc != 4)  {
		char *nl;	/* assume it's an error packet */
		free( player);
		server->error= malloc( pktlen+1);
	        nl= strchr( pkt, '\n');
		if ( nl != NULL)
		    strncpy( server->error, pkt, nl-pkt);
		else
		    strcpy( server->error, pkt);
		server->server_name= SERVERERROR;
		complete= 1;
		break;
	    }
	    player->connect_time *= 60;
	    pkt+= len;

	    if ( *pkt != '"') break;
	    pkt++;
	    end= strchr( pkt, '"');
	    if ( end == NULL) break;
	    player->name= (char*) malloc(end-pkt+1);
	    memcpy( player->name, pkt, end-pkt);
	    player->name[end-pkt]= '\0';
	    pkt= end+2;

	    if ( *pkt != '"') break;
	    pkt++;
	    end= strchr( pkt, '"');
	    if ( end == NULL) break;
	    player->skin= (char*) malloc(end-pkt+1);
	    memcpy( player->skin, pkt, end-pkt);
	    player->skin[end-pkt]= '\0';
	    pkt= end+2;

	    sscanf( pkt, "%d %d%n", &player->shirt_color, &player->pants_color,
		&len);
	    pkt+= len;

	    player->next= server->players;
	    server->players= player;
	    server->num_players++;
	}
	else
	    pkt++;
	complete= 1;
    }

    if ( !odd && !complete)
	print_packet( rawpkt, pktlen);

    cleanup_qserver( server, 0);
}

void
deal_with_q2_packet( struct qserver *server, char *rawpkt, int pktlen)
{
    char *key, *value, *end;
    struct rule *rule;
    struct player *player;
    int len, rc, odd= 0, complete= 0;
    char *pkt= &rawpkt[10];

/*	print_packet( rawpkt, pktlen); */
    rawpkt[pktlen]= '\0';

    while ( *pkt && pkt-rawpkt < pktlen)  {
	if ( *pkt == '\\')  {
	    pkt++;
	    end= strchr( pkt, '\\');
	    if ( end == NULL)
		break;
	    *end= '\0';
	    key= pkt;
	    pkt+= strlen(pkt)+1;
	    end= strchr( pkt, '\\');
	    if ( end == NULL)
		end= strchr( pkt, '\n');
	    value= (char*) malloc(end-pkt+1);
	    memcpy( value, pkt, end-pkt);
	    value[end-pkt]= '\0';
	    pkt= end;
	    if ( strcmp( key, "hostname") == 0)
		server->server_name= value;
	    else if  ( strcmp( key, "mapname") == 0 ||
		    (strcmp( key, "map") == 0 && server->map_name == NULL))  {
		if ( server->map_name != NULL)
		    free( server->map_name);
		server->map_name= value;
	    }
	    else if  ( strcmp( key, "maxclients") == 0)
		server->max_players= atoi(value);
	    else  {
		rule= (struct rule *) malloc( sizeof( struct rule));
		rule->name= strdup(key);
		rule->value= value;
		rule->next= server->rules;
		server->rules= rule;
		server->n_rules++;
	    }
	}
	else if ( *pkt == '\n')  {
	    pkt++;
	    if ( *pkt == '\0')
		break;
	    player= (struct player *) calloc( 1, sizeof( struct player));
	    rc= sscanf( pkt, "%d %d %n", &player->frags, &player->ping, &len);
	    player->number= -1;
	    player->connect_time= -1;
	    if ( rc != 2)  {
		char *nl;	/* assume it's an error packet */
		free( player);
		server->error= malloc( pktlen+1);
	        nl= strchr( pkt, '\n');
		if ( nl != NULL)
		    strncpy( server->error, pkt, nl-pkt);
		else
		    strcpy( server->error, pkt);
		server->server_name= SERVERERROR;
		complete= 1;
		break;
	    }
	    pkt+= len;

	    if ( *pkt != '"') break;
	    pkt++;
	    end= strchr( pkt, '"');
	    if ( end == NULL) break;
	    player->name= (char*) malloc(end-pkt+1);
	    memcpy( player->name, pkt, end-pkt);
	    player->name[end-pkt]= '\0';
	    pkt= end+1;

	    player->skin= "";
	    player->shirt_color= -1;
	    player->pants_color= -1;

	    player->next= server->players;
	    server->players= player;
	    server->num_players++;
	}
	else
	    pkt++;
	complete= 1;
    }

    if ( !odd && !complete)
	print_packet( rawpkt, pktlen);

    cleanup_qserver( server, 0);
}

/* Packet from QuakeWorld master server
 */
void
deal_with_qwmaster_packet( struct qserver *server, char *rawpkt, int pktlen)
{
    unsigned long ipaddr;
    unsigned short port;
    int i, offset= 2;

    server->ping_total+= time_delta( &packet_recv_time,
		&server->packet_time1);

    if ( rawpkt[0] == QW_NACK)  {
	server->error= strdup( &rawpkt[2]);
	server->server_name= SERVERERROR;
	cleanup_qserver( server, 1);
	return;
    }

    if ( *((int*)rawpkt) == -1)  {
	rawpkt+= 4;	/* QW 1.5 */
	pktlen-= 4;
	offset= 0;
    }

    if ( ! ( rawpkt[0] == QW_SERVERS  &&rawpkt[1] == QW_NEWLINE))  {
	fprintf( stderr, "Odd packet from QW master, processing anyway ...\n");
	print_packet( rawpkt, pktlen);
    }

    rawpkt[pktlen]= '\0';
    server->n_servers= 0;
    for ( i= 2; i < pktlen-offset; i+= 6)  {
	memcpy( &ipaddr, &rawpkt[i], 4);
	memcpy( &port, &rawpkt[i+4], 2);
	ipaddr= ntohl( ipaddr);
	port= ntohs( port);
	add_qserver_byaddr( ipaddr, port, QW_SERVER);
	server->n_servers++;
    }
    bind_sockets();
    server->server_name= MASTER;

    cleanup_qserver( server, 0);
}


int
server_info_packet( struct qserver *server, struct qpacket *pkt, int datalen)
{
    int off= 0;

    /* ignore duplicate packets */
    if ( server->server_name != NULL)
	return 0;

    server->address= strdup((char*)&pkt->data[off]);
    off+= strlen(server->address) + 1;
    if ( off >= datalen)
	return -1;

    server->server_name= strdup((char*)&pkt->data[off]);
    off+= strlen(server->server_name) + 1;
    if ( off >= datalen)
	return -1;

    server->map_name= strdup((char*)&pkt->data[off]);
    off+= strlen(server->map_name) + 1;
    if ( off > datalen)
	return -1;

    server->num_players= pkt->data[off++];
    server->max_players= pkt->data[off++];
    server->protocol_version= pkt->data[off++];

    server->retry1= n_retries;

    if ( get_server_rules)
	send_rule_request_packet( server);
    if ( get_player_info)
	send_player_request_packet( server);

    return 0;
}

int
player_info_packet( struct qserver *server, struct qpacket *pkt, int datalen)
{
    char *name, *address;
    int off, colors, frags, connect_time, player_number;
    struct player *player, *last;

    off= 0;
    player_number= pkt->data[off++];
    name= (char*) &pkt->data[off];
    off+= strlen(name)+1;
    if ( off >= datalen)
	return -1;

    colors= pkt->data[off+3];
    colors= (colors<<8) + pkt->data[off+2];
    colors= (colors<<8) + pkt->data[off+1];
    colors= (colors<<8) + pkt->data[off];
    off+= sizeof(colors);

    frags= pkt->data[off+3];
    frags= (frags<<8) + pkt->data[off+2];
    frags= (frags<<8) + pkt->data[off+1];
    frags= (frags<<8) + pkt->data[off];
    off+= sizeof(frags);

    connect_time= pkt->data[off+3];
    connect_time= (connect_time<<8) + pkt->data[off+2];
    connect_time= (connect_time<<8) + pkt->data[off+1];
    connect_time= (connect_time<<8) + pkt->data[off];
    off+= sizeof(connect_time);

    address= (char*) &pkt->data[off];
    off+= strlen(address)+1;
    if ( off > datalen)
	return -1;

    last= server->players;
    while ( last != NULL && last->next != NULL)  {
	if ( last->number == player_number)
	     return 0;
	last= last->next;
    }
    if ( last != NULL && last->number == player_number)
	return 0;

    player= (struct player *) malloc( sizeof(struct player));
    player->number= player_number;
    player->name= strdup( name);
    player->address= strdup( address);
    player->connect_time= connect_time;
    player->frags= frags;
    player->shirt_color= colors>>4;
    player->pants_color= colors&0xf;
    player->next= NULL;

    if ( last == NULL)
	server->players= player;
    else
	last->next= player;

    server->next_player_info++;
    server->retry2= n_retries;
    if ( server->next_player_info < server->num_players)
	send_player_request_packet( server);

    return 0;
}

int
rule_info_packet( struct qserver *server, struct qpacket *pkt, int datalen)
{
    int off= 0;
    struct rule *rule, *last;
    char *name, *value;

    /* Straggler packet after we've already given up fetching rules */
    if ( server->next_rule == NULL)
	return 0;

    if ( ntohs(pkt->length) == Q_HEADER_SIZE)  {
	server->next_rule= NULL;
	return 0;
    }

    name= (char*)&pkt->data[off];
    off+= strlen( name)+1;
    if ( off >= datalen)
	return -1;

    value= (char*)&pkt->data[off];
    off+= strlen( value)+1;
    if ( off > datalen)
	return -1;

    last= server->rules;
    while ( last != NULL && last->next != NULL)  {
	if ( strcmp( last->name, name) == 0)
	     return 0;
	last= last->next;
    }
    if ( last != NULL && strcmp( last->name, name) == 0)
	return 0;

    rule= (struct rule *) malloc( sizeof( struct rule));
    rule->name= strdup( name);
    rule->value= strdup( value);
    rule->next= NULL;

    if ( last == NULL)
	server->rules= rule;
    else
	last->next= rule;

    server->n_rules++;
    server->next_rule= rule->name;
    server->retry1= n_retries;
    send_rule_request_packet( server);

    return 0;
}


/* Misc utility functions
 */

char *
escape( unsigned char *string)
{
    static char _q[256];

    unsigned char *s= string;
    char *q= _q;

    if ( string == NULL)  {
	q[0]= '?';
	q[1]= '\0';
	return _q;
    }

    if ( hex_player_names)  {
	for ( ; *string; string++, q+= 2)
	    sprintf( q, "%02x", *string);
	*q= '\0';
	return _q;
    }

    for ( ; *string; string++)  {
	if ( isprint(*string))  {
	    *q++= *string;
	    continue;
	}

	if ( *string >= 0xa0)
	    *q++= *string & 0x7f;
	else if ( *string >= 0x92 && *string < 0x9c)
	    *q++= '0' + (*string - 0x92);
	else if ( *string >= 0x12 && *string < 0x1c)
	    *q++= '0' + (*string - 0x12);
	else if ( *string == 0x90 || *string == 0x10)
	    *q++= '[';
	else if ( *string == 0x91 || *string == 0x11)
	    *q++= ']';
	else if ( *string == 0xa || *string == 0xc || *string == 0xd)
	    *q++= ']';
    }
    *q= '\0';
    return _q;
}

int
is_default_rule( struct rule *rule)
{
    if ( strcmp( rule->name, "sv_maxspeed") == 0)
	return strcmp( rule->value, Q_DEFAULT_SV_MAXSPEED) == 0;
    if ( strcmp( rule->name, "sv_friction") == 0)
	return strcmp( rule->value, Q_DEFAULT_SV_FRICTION) == 0;
    if ( strcmp( rule->name, "sv_gravity") == 0)
	return strcmp( rule->value, Q_DEFAULT_SV_GRAVITY) == 0;
    if ( strcmp( rule->name, "noexit") == 0)
	return strcmp( rule->value, Q_DEFAULT_NOEXIT) == 0;
    if ( strcmp( rule->name, "teamplay") == 0)
	return strcmp( rule->value, Q_DEFAULT_TEAMPLAY) == 0;
    if ( strcmp( rule->name, "timelimit") == 0)
	return strcmp( rule->value, Q_DEFAULT_TIMELIMIT) == 0;
    if ( strcmp( rule->name, "fraglimit") == 0)
	return strcmp( rule->value, Q_DEFAULT_FRAGLIMIT) == 0;
    return 0;
}

char *
strherror( int h_err)
{
    static char msg[100];
    switch (h_err)  {
    case HOST_NOT_FOUND:	return "host not found";
    case TRY_AGAIN:		return "try again";
    case NO_RECOVERY:		return "no recovery";
    case NO_ADDRESS:		return "no address";
    default:	sprintf( msg, "%d", h_err); return msg;
    }
}

int
time_delta( struct timeval *later, struct timeval *past)
{
    if ( later->tv_usec < past->tv_usec)  {
	later->tv_sec--;
	later->tv_usec+= 1000000;
    }
    return (later->tv_sec - past->tv_sec) * 1000 +
	(later->tv_usec - past->tv_usec) / 1000;
}

int
connection_refused()
{
#ifdef unix
    return errno == ECONNREFUSED;
#endif

#ifdef _WIN32
    return WSAGetLastError() == WSAECONNABORTED;
#endif
}

void
print_packet( char *buf, int buflen)
{
    int i;
    for ( i= 0; i < buflen; i++)  {
	if ( isprint( buf[i])) fprintf( stderr, "%c", buf[i]);
	else fprintf( stderr, " %02x ", (unsigned char)buf[i]);
    }
    fprintf(stderr,"\n");
}

char *
quake_color( int color)
{
    static char *colors[] = {
	"White",	/* 0 */
	"Brown",	/* 1 */
	"Lavender",	/* 2 */
	"Khaki",	/* 3 */
	"Red",		/* 4 */
	"Lt Brown",	/* 5 */
	"Peach",	/* 6 */
	"Lt Peach",	/* 7 */
	"Purple",	/* 8 */
	"Dk Purple",	/* 9 */
	"Tan",		/* 10 */
	"Green",	/* 11 */
	"Yellow",	/* 12 */
	"Blue",		/* 13 */
	"Blue",		/* 14 */
	"Blue"		/* 15 */
    };

    if ( color_names)
        return colors[color&0xf];
    else
	return (char*)color;
}

char *
play_time( int seconds, int show_seconds)
{
    static char time_string[24];
    if ( time_format == CLOCK_TIME)  {
	time_string[0]= '\0';
	if ( seconds/3600)
	    sprintf( time_string, "%2dh", seconds/3600);
	else
	    strcat( time_string, "   ");
	if ( (seconds%3600)/60 || seconds/3600)
	    sprintf( time_string+strlen(time_string), "%2dm",
		(seconds%3600)/60);
	else if ( ! show_seconds)
	    sprintf( time_string+strlen(time_string), " 0m");
	else
	    strcat( time_string, "   ");
	if ( show_seconds)
	    sprintf( time_string+strlen(time_string), "%2ds", seconds%60);
    }
    else if ( time_format == STOPWATCH_TIME)  {
	if ( show_seconds)
	    sprintf( time_string, "%02d:%02d:%02d", seconds/3600,
		(seconds%3600)/60, seconds % 60);
	else
	    sprintf( time_string, "%02d:%02d", seconds/3600,
		(seconds%3600)/60);
    }
    else
	sprintf( time_string, "%d", seconds);

    return time_string;
}

char *
ping_time( int ms)
{
    static char time_string[24];
    if ( ms < 1000)
	sprintf( time_string, "%dms", ms);
    else if ( ms < 1000000)
	sprintf( time_string, "%ds", ms/1000);
    else
	sprintf( time_string, "%dm", ms/1000/60);
    return time_string;
}

void
add_to_array( char ***array, int *n, char *string)
{
    int i;
    if ( *array == NULL)  {
	*n= 10;
	*array= (char**) calloc( 1, sizeof(char*) * *n);
    }
    for ( i= 0; i < *n; i++)
	if ( (*array)[i] == NULL)  {
	    (*array)[i]= string;
	    return;
	}
    *n *= 2;
    *array= (char**) realloc( *array, sizeof(char*) * *n);
    memset( &(*array)[i], 0, sizeof(char*) * i);
    (*array)[i]= string;
}

int
count_bits( int n)
{
    int b= 0;
    for ( ; n; n>>=1)
	if ( n&1)
	    b++;
    return b;
}
